/*
RelCubeC : Complete information about the data cube
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package cube;

import java.util.*;
import java.io.*;
import java.sql.*;

import util.*;

public class RelCubeC extends FCubeC implements RelCube 
{
    String factTable;
    transient DB dbc;

    class RelDimInfo {
	String oidFldDet; // columname of this dimension in the fact table
	String dimTab; // name of the dimension table
	String hierOids[]; // oid column name of that level in the dimension table.
	String hierExtNames[]; // string name of that level in the dimension table.

	public RelDimInfo(DimInfo dim, BufferedReader din) throws Exception
	{
	    oidFldDet=din.readLine();
	    dimTab=din.readLine();
	    if (dimTab.equals("-"))
		return;
	    hierOids=new String[dim.nlevels];
	    hierExtNames=new String[dim.nlevels];
	    int i;
	    for(i=0;i<dim.nlevels;i++) {
		hierExtNames[i] = din.readLine(); // could be "-"	
		hierOids[i] = din.readLine(); // necessary
	    }
	}
	boolean namePresent(int lvl) {
	    if (dimTab.equals("-"))
		return false;
	    if (hierExtNames[lvl].equals("-"))
		return false;
	    return true;
	}
    };
    transient RelDimInfo rdims[];
    public RelCubeC(String ifile, String rfile, String userId, String passwd, boolean debug) throws Exception {
	super(ifile);
	BufferedReader din = new BufferedReader(new FileReader(rfile));
	String dbms = din.readLine();
	String dbName = din.readLine();
	if (dbms.equalsIgnoreCase("Oracle"))
	    dbc = new OracleDB(dbName, userId, passwd, debug);
	else
	    dbc = new DB2DB(dbName, userId, passwd, debug);
	factTable=din.readLine();
	rdims = new RelDimInfo[ndims];
	for (int i = 0; i < ndims; i++) {
	    rdims[i] = new RelDimInfo(dims[i], din);
	}
	din.close();
	initCubeStat(this);
    }
    public String relName() {return factTable;}
    public DB db() {return dbc;}
    public String dbName() {return dbc.dbName();}
    public String colName(int i, int j) {
	if (j > 0) 
	    return rdims[i].hierOids[j];
	else
	    return rdims[i].oidFldDet;
    }
    public String oidFld(int i, int j) {return rdims[i].hierOids[j];}
    String dimTable(int i) {return rdims[i].dimTab;}
    //
    // Given an attribute id, find the corresponding string name
    // of that attribute
    public String getExternalName(int dim, int lvl, Object colVal) {
	String mapstr = "";
	if (rdims[dim].namePresent(lvl)) {
	    mapstr += ("select distinct " + rdims[dim].hierExtNames[lvl] 
		       + " from " + rdims[dim].dimTab
		       + " where " + rdims[dim].hierOids[lvl] + " = " + colVal.toString());
	} else {
	    mapstr += ("char(" + colVal + ")");
	}
	return mapstr;
    }

    /*
    public String getNameFromId(int dim, int lvl, Object colVal)  throws Exception {
	String mapstr = "";
	if (rdims[dim].namePresent(lvl)) {
	    
	    return dbc.executeQuerySingleObject(getExternalName(dim,lvl,colVal)).toString();
	}
	return colVal.toString();
    }
    */
    //
    // Given an attribute id and string name, find the corresponding 
    // integer oid of that attribute value
    //
    public Object memberIdFromName(int dim, int lvl, String colVal) throws DataCubeException {
	if (rdims[dim].namePresent(lvl)) {
	    String qry = ("select distinct " + rdims[dim].hierOids[lvl] 
			  + " from " + rdims[dim].dimTab
			  + " where " + rdims[dim].hierExtNames[lvl] + " = '" + colVal.trim() + "'");
	    try {
		return dbc.executeQuerySingleInteger(qry);
	    } catch  (Exception e) {
		throw new DataCubeException("Invalid dimension value: " + colVal + " " + e.getMessage());
	    } 
	} else {
	    return new Integer(colVal);
	}
    }
    public String subCubeQuery(Term t, String meas) {
	return subCubeQuery(t,meas,null);
    }
    
    public String subCubeQuery(Term t, String desiredMeasures, Vector exclTerms) {
	return subCubeQuery(t, desiredMeasures, exclTerms,null);
    }
    
    // these are to avoid repeated new
    // 
    private SqlString grpList = new SqlString(", ");
    private SqlString sortList = new SqlString(", ");
    private SqlString whereList = new SqlString(" AND ");
    private boolean dimJoin[]; // a temporary array.
    private String tmpcols[]; // a temporary array
    
    public String subCubeQuery(Term t, String desiredMeasures, Vector exclTerms, int[] finalSort) {
	return subCubeQuery(t, desiredMeasures, exclTerms, finalSort, factTable);
    }

    String sortClause(Term t, int[] finalSort) {
	if (finalSort != null) { 
	    sortList.clear();
	    for (int i = 0; (i < finalSort.length) && (finalSort[i] >= 0); i++) {
		sortList.add(t.desc(finalSort[i]).name());
	    }
 	    return sortList.str(" order by ");
	}
	return "";
    }
    String addSortClause(String st, Term t, int[] finalSort) {
	return st + sortClause(t,finalSort);
    }
    public String subCubeQuery(Term t, String desiredMeasures, Vector exclTerms, int[] finalSort,  String sourceTable) {
	String projList = ""; 
	String relList = "";
	grpList.clear();
	whereList.clear();
	
	if (dimJoin == null)
	    dimJoin = new boolean[ndims];
	if ((tmpcols == null) || (tmpcols.length < t.numAttrs()))
	    tmpcols = new String[t.numAttrs()];

	for (int i = 0; i < ndims; i++) {
	    dimJoin[i] = false;
	}
	for (int i = 0; i < t.numAttrs(); i++) {
	    String col = colName(t.desc(i).dim(), t.desc(i).level());
	    if (t.desc(i).level() == 0) {
		col = "f." + col;
	    } else {
		int d = t.desc(i).dim();
		col = "d" + d + "." + col;
		dimJoin[d] = (dimJoin[d] || !t.aggregated(i));
	    }
	    tmpcols[i] = col;
	    if (tmpcols[i].equals("f.grpoid")) {
		System.out.println("WARNING: " + i + t.toString());
	    }
	    if (t.range(i)) {
		projList += (col + " ");
		grpList.add(col);
	    } else if (t.constant(i)) {
		projList += (t.val(i) + " ");
		whereList.add(" ("+col+  " = " + t.val(i)+") "); 
	    } else {
		projList += (Term.AGGREGATED + " ");
	    }
	    projList += "as " + t.desc(i).name() +", ";
	}

	if (exclTerms != null) {
	    for (int i = 0; i < exclTerms.size(); i++) {
		Term aterm = (Term)exclTerms.elementAt(i);
		for (int j = 0; j < aterm.numAttrs(); j++) {
		    if (dimJoin[aterm.desc(j).dim()])
			continue;
		    if (aterm.constant(j) && (aterm.desc(j).level() > 0))
			dimJoin[aterm.desc(j).dim()] = true;
		}
	    }
	}
	for (int d = 0; d < numDims(); d++) { 
	    if (dimJoin[d]) {
		relList += (", " + rdims[d].dimTab + " d" + d);
		whereList.add(" (f." + rdims[d].oidFldDet + " = "  
			      + "d" + d + "." + rdims[d].hierOids[0]+ ") ");
	    }
	}
	if (exclTerms != null) {
	    for (int i = 0; i < exclTerms.size(); i++) {
		whereList.add(" not (" + SqlQryGen.whereClause(tmpcols, (Term)exclTerms.elementAt(i)).str("")+")");
	    }
	}
	String st = ("select " + projList 
		     + desiredMeasures
		     + " from " + sourceTable + " f " + relList 
		     +  whereList.str(" where ")
		     +  grpList.str(" group by ")
		     );	
	return addSortClause(st, t, finalSort);
    }

    public String mergedSubCubeQuery(Term t1, Term t2) throws SQLException
    {
	return mergedSubCubeQuery(t1,t2,false);
    }
    public String mergedSubCubeQuery(Term t1, Term t2, boolean orderResult) throws SQLException
    {
	String st1 = subCubeQuery(t1, "sum("+measure(0)+") as v1, 0 as v2");
	String st2 = subCubeQuery(t2, "0 as v1, sum("+measure(0)+") as v2");
	// TODO.
	SqlString qry = SqlQryGen.groupList(t1);
	String st = "select " 
	    + SqlQryGen.projectNames(t1)
	    + "sum(v1) as v1, sum(v2) as v2"
	    + " from (" + st1 + " union " + st2 + ") t" 
	    + qry.str(" group by ");
	if (orderResult) st += qry.str(" order by ");
	return st;
    }
    /*
    public Iter mergedSubCube(Term t1, Term t2, TermDesc tdesc) throws Exception
    {
	return new RelDataIter(dbc.executeQuery(mergedSubCubeQuery(t1, t2,true)), tdesc);
	} 
    public Iter mergedSubCube(Term t1, Term t2, TermDesc tdesc, int order[]) throws Exception
    {
	String st = mergedSubCubeQuery(t1, t2,false);
	sortList.clear();
	for (int i = 0; i < order.length; i++) {
	    sortList.add(tdesc.desc(order[i]).name());
	}
	st += sortList.str(" order by ");
	return new RelDataIter(dbc.executeQuery(st),tdesc);
	}*/
    public ResultSet subCube(Term t) throws SQLException {
	return  subCube(t,false,0);
    }
    public ResultSet subCube(Term t, boolean sorted) throws SQLException {
	return subCube(t,sorted, 0);
    }
    public ResultSet subCube(Term t, boolean sorted, int measId) throws SQLException {
	return dbc.executeQuery(subCubeQuery(t, "sum("+measure(measId)+")", null, (sorted==false)?null:MdUtils.defaultArray(t)));
    }

    public Object getParent(int  childDim, int childLvl, Object id, int parLvl) throws Exception {
	return dbc.executeQuerySingleInteger("select distinct " + oidFld(childDim, parLvl) + " from " + dimTable(childDim) + " where " + oidFld(childDim, childLvl) + " = " + id.toString());
    }
    
    void readMembers(HierStat hier, ResultSet rs, Vector tempVec) {
	try {
	    tempVec.removeAllElements();
	    int numCols = rs.getMetaData().getColumnCount();
	    while (rs.next()) {
		for (int i = 0; i < numCols; i++) {
		    tempVec.addElement(rs.getObject(i+1));
		}
	    } 
	    DB.cursorClose(rs);
	    hier.members = new Object[tempVec.size()/numCols];
	    hier.names = new String[tempVec.size()/numCols];
	    int vind = 0;
	    for (int i = 0; i < hier.members.length; i++) {
		hier.members[i] = tempVec.elementAt(vind++);
		if (numCols > 1) {
		    hier.names[i] = (String)tempVec.elementAt(vind++);
		} else {
		    hier.names[i] = tempVec.elementAt(i).toString();
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }
    void initCubeStat(RelCubeC c) {
	c.initCubeStat();
	Vector tempVec = new Vector();
	for (int i = 0; i < c.numDims(); i++) {
	    ResultSet rs;
	    try {
		if (c.dimTable(i).equals("-")) {
		    rs = c.db().executeQuery("select distinct " + c.colName(i,0)
					     + " from " + c.relName());
		    readMembers(c.dimStat[i].hierStat[0], rs, tempVec);
		    continue;
		} 
		
		for (int j = 0; j < c.numLevels(i); j++) {
		    String st = "select distinct "+ c.oidFld(i,j);
		    if (rdims[i].namePresent(j))
			st += (", " + c.rdims[i].hierExtNames[j]);		    
		    rs = c.db().executeQuery(st + " from " + c.dimTable(i));
		    readMembers(c.dimStat[i].hierStat[j], rs, tempVec);
		}
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    public QueryHandle subCube(Term t, 
			       TermDesc projCols, 
			       MeasureExpr[] measExpr, 
			       int[] sortOrder) throws Exception
    {
	String st = subCubeQuery(t, toSqlString(measExpr).str(""), null, sortOrder);
	return new RelQueryHandle(st,t.termDescr());
    }
    public CubeRows executeQuery(QueryHandle q) throws Exception {
	RelQueryHandle rq = (RelQueryHandle)q;
	return new RelCubeRows(dbc.executeQuery(rq.queryString()), rq.projColumns());
    }
    SqlString toSqlString(MeasureExpr meas[]) {
	mapTo(meas);
	SqlString measStr = new SqlString(","); 
	for (int i = 0; i < meas.length; i++)
	    measStr.add(meas[i].toString() + " as " + meas[i].name());
	return measStr;
    }
    void mapTo(MeasureExpr meas[]) {
	if (dbc.dbType().equals("Oracle")) return;
	else if (dbc.dbType().equals("DB2")) {
	   for (int i=0; i<meas.length; i++) {
	      String s = meas[i].toString();
	      int k = s.indexOf("NVL"); 
	      StringBuffer sb = new StringBuffer(s);
	      while (k!=-1) {		
	        sb.replace(k,k+3,"coalesce");
		s = sb.toString();  
		k = s.indexOf("NVL");
	      }	 	
	      meas[i].setExpr(sb.toString());
	   }
 	}
    }	
    String newMeasureName(int i) {
	return "v" + i;
    }
    void toSqlString(MeasureExpr m[], int startIndex, SqlString str) {
	mapTo(m);
	for (int i = 0; i < m.length; i++)
	    str.add(m[i].toString() + " as " + newMeasureName(startIndex+i));
    }
    void zeroMeasures(int len, int startIndex, SqlString meas) {
	for (int i = 0; i < len; i++)
	    meas.add("0 as " + newMeasureName(startIndex+i));
    }
    
    public QueryHandle concatCubes(Term terms[], 
				   TermDesc tdesc, 
				   MeasureExpr mexprs[][], 
				   int[] sortOrder)  throws Exception
    {
	int totalMeas = 0;
	SqlString unionQry = new SqlString(" union ");
	for (int p = 0; p < terms.length; p++) {
	    SqlString thisMeasStr = new SqlString(",");
	    int prefixMeas = 0;
	    for (int m1 = 0; m1 < terms.length; m1++) {
		if (m1 != p) {
		    zeroMeasures(mexprs[m1].length, prefixMeas, thisMeasStr);
		} else {
		    toSqlString(mexprs[p], prefixMeas, thisMeasStr);
		}
		prefixMeas += mexprs[m1].length;
	    }
	    unionQry.add(subCubeQuery(terms[p], thisMeasStr.str(""))); 
	    
	    totalMeas += mexprs[p].length;
	}
	SqlString measStr = new SqlString(",");
	for (int i = 0; i < totalMeas; i++) {
	    String measName = newMeasureName(i);
	    measStr.add("sum(" + measName + ") as " + measName);
	}
	SqlString qry = SqlQryGen.groupList(terms[0]);
	String st = "select " 
	    + SqlQryGen.projectNames(tdesc)
	    + measStr.str("")
	    + " from (" + unionQry.str("") + ") t " 
	    + qry.str(" group by ");
	return new RelQueryHandle(addSortClause(st, terms[0], sortOrder),tdesc);
    }
    public QueryHandle subCube(Term t, 
			       TermDesc projCols, 
			       MeasureExpr measExpr[], 
			       QueryHandle q, 
			       int[] sortOrder) throws Exception
    {
	RelQueryHandle rq = (RelQueryHandle)q;
	String st = "select ";
	if (projCols!=null) st+= SqlQryGen.projectNames(projCols);
	st+= toSqlString(measExpr).str("")
	    + " from ("+rq.queryString()+") y "
	    + SqlQryGen.whereClause(t,false).str(" where ") // TODO -- need to check against proj cols
	    + SqlQryGen.groupList(t).str(" group by ")
	    + sortClause(t, sortOrder);
	return new RelQueryHandle(st, projCols);
    }
}



