/*
FindDiff : Finding differences 
    Copyright (C) 2001 Sunita Sarawagi

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package diff;

import java.util.*;

import util.*;


public class FindDiff {
    Tuple EmptyTup;
    public static final float TOLERANCE = (float)0.1;
    boolean debug = false;

    public interface NNode extends Node {
	int maxNum();
	double cost(int k, double r);
	// part of cost not affected by the external ratio r.
	double coveredCost(int k, double r);
	Tuple aggTup(int k, double r);
	Tuple tup(int k, double r, int i);
	void setRVals(RVals rvals);
    };
    public class NTupleNode implements  NNode
    {
	// tuple currently representing this node..
	//aggregates measures of everything for which it is parent
	Tuple _currTup = new Tuple(); 

	public Tuple tuple() {return _currTup;}

	double cst0;
	double cst1;
	double prevR;
	public void init(Tuple tup) {
	    _currTup.copy(tup); cst1 = _currTup.cost();prevR=-2;
	} 
	public int maxNum() {return 1;}
	public void setRVals(RVals rvals) {;}	
	public double cost(int k, double r) {
	    if (k == 0) {
		if (prevR == r)
		    return cst0;
		prevR = r;
		cst0 =  _currTup.error(r);
		//      MyOut.println(_currTup + " " + _currTup.error(r) + endl;
		return cst0;
	    }
	    return cst1;
	}
	public double coveredCost(int k, double r) 
	{return (k==0)?0:cst1;}
	public Tuple aggTup(int k, double r)  { 
	    if (k == 1) return _currTup; else return EmptyTup;
	}
	public Tuple tup(int k, double r, int i)  {
	    // assert(i == 0); assert(_currTup.ndim() > 0); 
	    return _currTup;
	}
    };
    
    class Soln implements DiffResult {
	int sz = 0;
	double _cst = 0;
	double _coveredCost;
	Tuple _tups[]; // array of tuple pointer
	Tuple _aggTup; // agg of all tuples included in the solution list, does not
	Tuple tmp;

	public Soln(int size,double cst){init(size,cst);}
	public void init(int size,double cst) {
	    sz = size; _cst = cst; _coveredCost = cst; 
	    _tups = null; 
	    if (sz > 0) 
		_tups = new Tuple[sz];
	    _aggTup = new Tuple();
	    for (int i = 0; i < sz; i++)
		_tups[i] = new Tuple();
	
	    tmp = new Tuple();
	}

	public void start() {_cst = _coveredCost = 0; sz = 0; _aggTup.initMeasures();} 
	// cost not affected by outside ratio..
	public double coveredCost() {return _coveredCost;} 

	public double cost() { return _cst;}

	public int size() {return sz;}

	public Tuple aggTup() {return _aggTup;}
	public int numTups() {return size();}
	public Tuple tup(int i) {
	    return _tups[i];
	}
	public float error(int i) {
	    return (float)tuple(i).error();
	}
	public Tuple tuple(int i) {
	    return _tups[i];
	}
	public void updateSoln(double cst, double r, Soln sol1, int sz2, NNode sol2) {
	    int sz1 = sol1.size();
	    _cst = cst;
	    _coveredCost = sol1.coveredCost() + sol2.coveredCost(sz2,r);
	    _aggTup.copyMeasures(sol1.aggTup());
	    _aggTup.addChild(sol2.aggTup(sz2, r));
	    if (sol1 == this)
		return;
	    for (int i = 0; i < sz1; i++)
		_tups[i].copy(sol1.tuple(i));
	    for (int j = 0; j < sz2; j++)
		_tups[sz1+j].copy(sol2.tup(sz2, r, j));
	    sz = sz1+sz2;
	}
	public void updateSoln(double cst, Soln sol1, Tuple par) {
	    _cst = cst;
	    _coveredCost = cst;
	    _aggTup.copyMeasures(par);
	    int sz1 = sol1.size();
	    sz = sz1+1;
	    _tups[0].copy(par);
	    _tups[0].removeChild(sol1.aggTup());
	    // assert(_tups[0].ndim() > 0);
	    for (int i = 0; i < sz1; i++)
		_tups[i+1].copy(sol1.tuple(i));
	}
	public void updateCoveredSoln(double cst, Soln sol1, Tuple par) {
	    _cst = cst;
	    _coveredCost = cst;
	    _aggTup.copyMeasures(par);
	    sz = sol1.size();
	    for (int i = 0; i < sol1.size(); i++)
		_tups[i].copy(sol1.tuple(i));
	}
	public  double actualCost(double rp, Soln sol1, Tuple par) {
	    tmp.copyMeasures(par);
	    tmp.removeChild(sol1.aggTup());
	    // if actual R close to the r for which children's cost is calculated
	    // return cached cost..
	    //    if (absVal(m.ratio() - rp) < TOLERANCE)
	    //  return sol1.cost();
	    // otherwise recompute cost.
	    return sol1.coveredCost() + tmp.cost();
	}
	public void updateCost(Tuple par, double newR) {
	    _cst = actualCost(par,newR); 
	}
	public double actualCost(Tuple par, double newR) {
	    if (coveredCost() == cost())
		return cost();
	    tmp.copyMeasures(par);
	    tmp.removeChild(aggTup());
	    if (tmp.count() < 0) 
		output();
	    return coveredCost() + tmp.error(newR);
	}

	void output(double finalR) {
	    MyOut.println(" size " + size() + " cost " + cost());
	    for (int i = 0; i < size(); i++) {
		MyOut.println( tup(i).toString() + " " + ((Tuple)tup(i)).error(finalR));
		//      MyOut.println( tup(i)._dvals[0] + " " + tup(i)._dvals[1] + " " + tup(i)._dvals[2] + " " + tup(i)._dvals[3] );
	    }
	}
	public void output() {
	    output(_aggTup.ratio());
	}
    };

    Soln InFeasibleSoln = new Soln(0,Double.MAX_VALUE);
    public class INNode implements INode, NNode {
	int _N; // maximum number of slots.
	int _maxNum; // current maximum slots <= _N
	
	// tuple currently representing this node..
	//aggregates measures of everything for which it is parent
	protected Tuple _currTup = new Tuple(); 
	public Tuple tuple() {return _currTup;}

	// cost arrays
	Soln _soln[][];
	boolean rootNode;
	int minRind; // ratio correspoinding to smallest r for maxNum
	double _prevR;
	int prevInd;
	boolean covered = false;

	RVals _rval;
	public  INNode(DiffParams pr) {
	    _soln = null;	    
	    _rval = new RVals();
	    init(pr);
	}
	public void init(Tuple t) {;}
	public void setAsRoot() {rootNode = true;}
	public void init(DiffParams pr){
	    rootNode = false;
	    _N = pr.N(); 
	    _rval.init(pr.ratio(), pr.maxR(), pr.ratios());
	    _maxNum = 0;
	    _soln = new Soln[_rval.size()][_N+1];
	    for (int i = 0; i < _rval.size(); i++) {
		for (int k = 0; k <= _N; k++) {
		    _soln[i][k] = new Soln(k,0);
		}
	    }
	}
	public int maxNum() {return _maxNum;}
	public double coveredCost(int k, double r) {return soln(k,r).coveredCost();}
	public double cost(int k, double r)  //best cost for k slots and external ratio r
	{ 
	    if (_prevR == r) 
		return soln(k,prevInd).cost();
	    _prevR = r;
	    prevInd = _rval.closestRInd(r);
	    if (Math.abs(r-_rval.val(prevInd))/_rval.val(prevInd) < TOLERANCE)
		return soln(k,prevInd).cost();
	    return soln(k,prevInd).actualCost(_currTup,r);
	}
	public double cost(int k, int i) //best cost for k slots and external ratio r
	{    return soln(k,i).cost();  }
	public Soln soln(int k, double r) {
	    return _soln[_rval.closestRInd(r)][k];
	}
	public Soln soln(int k, int i) {
	    return _soln[i][k];
	}
	public Tuple aggTup(int k, double r)  
	{return soln(k,r).aggTup();}
	public Tuple tup(int k, double r, int i)  
	{return (Tuple)soln(k,r).tup(i);}
  
	public void setRVals(RVals rvals) {_rval.setRVals(rvals);}
	public void start(Node cNode, BitSet id) // initialize starting costs
	{
	    NNode cnode = (NNode)cNode;
	    _maxNum = 0; 
	    _prevR = -1;
	    ((NNode)cnode).setRVals(_rval);
	    _currTup.copy(cnode.tuple());
	    _currTup.mask(id);
	    _currTup.initMeasures(); 
	    // re-init the r-values
	    for (int i = 0; i < _rval.size(); i++) {
		for (int k = 0; k <= _N; k++)
		    _soln[i][k].start();
	    }
	    add(cnode);
	    
	    // System.out.println(_currTup.toString());
	    // _rval.output();
	}
	int prevIndex(int i, int pn)
	{
	    return Math.min(i,pn);
	}
	public void add(Node cNode)  // include new node.
	{
	    NNode cnode = (NNode)cNode;
	    //    calcRVal(cnode.tuple());
	    _currTup.addChild(cnode.tuple());
	    int pN = maxNum();
	    int n = cnode.maxNum();
	    int N = Math.min(_N, pN+n);
	    // find best solution excluding the parent node.
	    int rind = 0;
	    _maxNum = Math.max(N, _maxNum);
	    int lastRIndex = _rval.size()-1;
	    int prevSolInd;
	    //
	    // for the last R, calculate the best value over all r-s
	    //
	    if (! _rval.externalRatio()) {
	    int ri = _rval.size()-1;
	    for (int j = N; j >= 0; j--) {
		double mincost = Double.MAX_VALUE;
		int cn = Math.min(n, j);
		int mini = 0;
		double minR = 0;
		for (int i = 0; i <= cn; i++) {
		    // find best r as the ratio due to the aggTup -
		    // explicit tuples in the merge of these two
		    // solutions.
		    double bestR = _rval.getBestR(_currTup, soln(prevIndex(j-i, pN), ri).aggTup(), cnode.aggTup(i,_rval.val(ri))); 
		    double cst = cost(prevIndex(j-i,pN), bestR) + cnode.cost(i, bestR);
		    if (cst >= mincost)
			continue;
		    mincost = cst;
		    mini = i;
		    minR = bestR;
		}
		soln(j,ri).updateSoln(mincost, minR, soln(prevIndex(j-mini,pN), minR), mini, cnode);
		if ((j == N-1) && (minR != 0)) {
		    _rval.update(minR, ri);
		    // Propagate this rval to the child node for better guess on rval
		    // for next tuple..
		    cnode.setRVals(_rval);
		}
	    }
	    } else {
		lastRIndex = _rval.size();
	    }
	    for (double r = _rval.val(0); rind < lastRIndex; r = _rval.val(++rind)) {
		for (int j = N; j >= 0; j--) {
		    double mincost = Double.MAX_VALUE;
		    int cn = Math.min(n, j);
		    int mini = 0;
		    for (int i = 0;  i <= cn; i++) {
			double cst = cost(prevIndex(j-i,pN), rind) + cnode.cost(i, r);
			if (cst >= mincost)
			    continue;
			mincost = cst;
			mini = i;
		    }
		    soln(j,rind).updateSoln(mincost, r, soln(prevIndex(j-mini,pN), rind), mini, cnode);
		}
	    }
	}
	public void close() {
	    if (covered) {
		coveredClose();
		return;
	    }
	    // having gotten all the tuples, correct some r-values if necessary
	    // checkUpdateRatios();
	    // For each r-value and j, update the cost to be the minimum of the cost with or 
	    // without the tuple at that node included.

	    if (debug)
		System.out.println("Closing solution for node: " + tuple());
	    for (int j = maxNum(); j >= 1; j--) {

		// Find the minimum cost if the root node at this level is included.
		double mincost = Double.MAX_VALUE;
		double minR = 0;
		int rind = 0;
		int minRi = 0;
		int minsz = Integer.MAX_VALUE;
		for (double rp = _rval.val(0); rp != -1; rp = _rval.val(++rind)) {
		    double cst = InFeasibleSoln.actualCost(rp, soln(j-1,rind), _currTup);
		    if (debug)
			System.out.println("Size " + j + " r " + rp + " cost " + cst);
		    if (cst > mincost)
			continue;
		    if (soln(j-1,rind).size() > minsz)
			continue;
		    mincost = cst;
		    minR = rp;
		    minRi = rind;
		    minsz = soln(j-1,rind).size();
		    if (rootNode && _rval.externalRatio())
			break;
		}
		if (debug) {
		    _rval.output();
		    MyOut.println( "Ratio " + minR + " old cost " + cost(j-1, minRi));
		    soln(j-1, minRi).output(minR);
		    if (minRi != _rval.size()-1) {
			MyOut.println( "Answer different from best calculated " );
			soln(j-1, _rval.size()-1).output();
		    }
		}
		if (rootNode) {
		    soln(j,minRi).updateSoln(mincost, soln(j-1, minRi), _currTup);
		    if (j == maxNum())
			minRind = minRi;
		    continue;
		} 
		int ri = 0;
		for (double r = _rval.val(0); r != -1; r = _rval.val(++ri)) {
		    if (cost(j,ri) > mincost) {
			// assert(mincost != Double.MAX_VALUE);
			soln(j,ri).updateSoln(mincost, soln(j-1,minRi), _currTup);
		    }
		}
	    }
	    /*
	      double mincost = Double.MAX_VALUE;
	      int rind = 0;
	      for (double rp = rval(0); rp != -1; rp = rval(++rind)) {
	      if (mincost > cost(maxNum(), rind)) {
	      mincost = cost(maxNum(), rind);
	      minRind = rind;
	      }
	      }
	    */
	}
	public void setCoveredStatus(boolean status) {
	    covered = status;
	}
	// this corresponds to internal node whose tuples are known to the
	// user, hence can be excluded from consideration.
	public void coveredClose() {
	for (int j = maxNum(); j >= 1; j--) {
	    double mincost = Double.MAX_VALUE;
	    double minR = 0;
	    int rind = 0;
	    int minRi = 0;
	    for (double rp = _rval.val(0); rp != -1; rp = _rval.val(++rind)) {
		double cst = InFeasibleSoln.actualCost(rp, soln(j,rind), _currTup);
		if (cst >= mincost)
		    continue;
		mincost = cst;
		minR = rp;
		minRi = rind;
	    }
	    rind = 0;
	    for (double rp = _rval.val(0); rp != -1; rp = _rval.val(++rind)) {
		soln(j,rind).updateCoveredSoln(mincost, soln(j, minRi), 
					       _currTup);
	    }
	}  
	}
	public DiffResult finalResult() {
	    if (debug) {
		MyOut.println( "Final solution index " + minRind + " " + maxNum() );
		MyOut.println( "Min Cost " + _soln[minRind][maxNum()].cost());
	    }
	    return _soln[minRind][maxNum()];
	}
	public DiffResult getResult(int numSlots) {
	    if (numSlots > maxNum())
		numSlots = maxNum();
	    return _soln[minRind][numSlots];
	}
	
	/*  void checkUpdateRatios() {
	    // check for own ratio
	    double r = _currTup.ratio();
	    int repl = checkRatios(r);
	    if (repl > 0) 
	    updateRatio(repl, r);
	    // check for ratio after removing some outlying children
	    }

	    int checkRatios(double r) {
	    // find first and second closest values to r
	    double close1 = Double.MAX_VALUE;
	    double close2 = Double.MAX_VALUE; 
	    double ind1;
	    double ind2;
	    for (int i = 0; i < _numR; i++) {
	    if (absVal(close1-r) > absVal(_rvals[i]-r)) {
	    close2 = close1;
	    ind2 = ind1;
	    close1 = _rvals[i];
	    ind1 = i;
	    continue;
	    }
	    if (absVal(close2-r) > absVal(_rvals[i]-r)) {
	    ind2 = i;
	    close2 = _rvals[i];
	    }
	    }
	    assert((close1 != Double.MAX_VALUE)  (close2 != Double.MAX_VALUE));
	    for (int k = 0; k < maxNum(); k++) {
	    double projectedCostDiff = absVal(cost(k,ind1) - cost(k,ind2))*absVal(close1-r)/absVal(close1-close2);
	    if (cost(k,ind1) < Double.MIN_VALUE) continue;
	    if (projectedCostDiff/cost(k,ind1) < 0.1)
	    continue;
	    return ind1;
	    }
	    return -1;
	    }
	    void updateRatio(int rind, double newR) {
	    for (int k = 0; k < maxNum(); k++)
	    _soln[rind][k].updateCost(_currTup, newR);
	    _rvals[rind] = newR;
	    } */
    };

    class Agg {
	public  CoverStatus _status = null;
	Node nextNode() {return null;}
	void setAsRoot() {;}
	Agg linkNode() {return null;}
	public Agg(CoverStatus status) {_status = status;}
	public boolean covered(Tuple tup) {
	    if (_status != null)
		return _status.covered(tup);
	    else 
		return false;
	}
    };

    class DBAgg extends Agg {
	Node _currNode;
	Iter _db;
	Tuple tup;
	
	DBAgg(Iter iter, CoverStatus status, DiffParams pr)
	{
	    super(status);
	    _db = iter;
	    if (pr.fixedN()) {
		_currNode = new NTupleNode();
	    } else {
		_currNode = new ETupleNode(pr);
	    }
	    tup = new Tuple(iter.termDesc());
	}
	Node nextNode()
	{
	    if (! _db.moreTuples()) {
		_db.close();
		return null;
	    }
	    _db.getTuple(tup);
	    while (covered(tup) && _db.moreTuples()) {
		_db.getTuple(tup);		
	    }
	    tup.calcDerivedCols();
	    _currNode.init(tup);
	    if (debug) 
		MyOut.println("R " + _currNode.tuple() );
	    return _currNode;
	}
    };

    class IAgg extends Agg {
	BitSet _id;
	Agg _link;
	Node _currChild;
	INode _currNode;

	IAgg(Agg linkNode, BitSet id, DiffParams pr, CoverStatus status)
	{
	    super(status);
	    _link = linkNode;
	    _id = new BitSet(id.size());
	    _id.or(id);
	    _currChild = _link.nextNode();
	    if (pr.fixedN()) {
		_currNode = new INNode(pr);
	    } else
		_currNode = new IENode(pr);
	}
	void output() {MyOut.println( _id);}
	BitSet id() {return _id;}
	void setAsRoot() {_currNode.setAsRoot();}
	Node nextNode() 
	{
	    if (_currChild == null) return null;
	    _currNode.start(_currChild, _id);
	    _currNode.setCoveredStatus(covered(_currNode.tuple()));
	    while ((_currChild = _link.nextNode()) != null) {
		if (! _currNode.tuple().isEqual(_currChild.tuple(), _id))
		    break;
		_currNode.add(_currChild);
	    }
	    if (debug) MyOut.println("I "+ _currNode.tuple() );
	    _currNode.close();
	    return _currNode;
	}
	Agg linkNode() {return _link;}
	DiffResult finalResult() {return _currNode.finalResult();}
	AllDiffResult allFinalResults() {return _currNode;}
    };


    //
    // Given an iterator on a stream of tuples with two values, this class finds
    // the summarized difference between the values using at most N tuples (specified
    // by the DiffParams field). 
    //
    IAgg lastAgg = null;
    
    DiffResult finalRes;
    AllDiffResult allFinalRes;
    void getDiff(Iter iter, DiffParams p, CoverStatus status, int[] levelOrder) {
	Tuple.setErrorFunction(p.errorFunction());
	EmptyTup = new Tuple();
	Agg ag = new DBAgg(iter, status, p);
	int numLevels = iter.numAttrs();
	//	assert(iter.numLevels() < 20);
	debug = p.debug();
	
	BitSet id = new BitSet(numLevels);
	for (int i = 0; i < numLevels; id.set(i++));
	for (int i = 1; i <= numLevels; i++) {
	    if (levelOrder == null)
		id.clear(numLevels-i);
	    else
		id.clear(levelOrder[numLevels-i]);
	    if (debug) MyOut.println( id );
	    lastAgg = new IAgg(ag,id,p, status);
	    ag = lastAgg;
	}
	lastAgg.setAsRoot();
	lastAgg.nextNode(); 

    }
    //
    // TODO -- make reuse of space possible..
    public FindDiff(Iter iter, DiffParams p, CoverStatus status, int[] levelOrder) 
    {
	getDiff(iter, p, status, levelOrder);
	finalRes = lastAgg.finalResult();
	allFinalRes = lastAgg.allFinalResults();
    }
    

    public DiffResult finalResult() {return finalRes;}
    public AllDiffResult allFinalResults() {return allFinalRes;}
    int random(int sz) {
	return ((int)Math.rint(Math.random()*sz)) % sz;
    }
    void randomPermutation(int ndims, int[] dimOrder) {
	for (int i = 0; i < ndims; i++) {
	    int pos = random(ndims-i) + i;
	    int tmp = dimOrder[i];
	    dimOrder[i] = dimOrder[pos];
	    dimOrder[pos] = tmp;
	}
    }
    /*
    public int numAttrs() {
	return tdesc.size();
    }
    public int numDims() {
	int ndims = 0;
	int thisDim = -1;
	for (int i = 0; i < tdesc.size(); i++) {
	    if (tdesc.desc(i).dim() != thisDim) {
		thisDim = tdesc.desc(i).dim();
		ndims++;
	    }
	}
	return ndims;
    }
    public int numLevels(int dim) {
	int nlvls = 0;
	for (int i = 0; i < tdesc.size(); i++) {
	    if (tdesc.desc(i).dim() == dim) 
		nlvls++;
	}
	return nlvls;
    }
    boolean getNewOrder(int[] levelOrder, TermDesc tdesc) {
	int ndims = itergen.numDims();
	int dimOrder[] = new int[ndims];
	for (int i = 0; i < ndims; dimOrder[i] = i, i++);
	    // pick a random order
	randomPermutation(ndims, dimOrder);
	for (int d = 0,r=0; d < ndims; d++) {
	    int dim = dimOrder[d];
		levelOrder[r++] = itergen.dimIndex(dim);
		for (int l = itergen.numLevels(dim)-2; l >= 0; l--)
		    levelOrder[r++] = levelOrder[r-1]+1;
	    }
    }
    */

    boolean getNewOrder(int[] levelOrder, TermDesc tdesc) {
	if (levelOrder[0] == -1) {
	    for (int i = 0; i < levelOrder.length; 
		 levelOrder[i] = i, i++);
	    return true;
	}

	boolean retval = getNewOrder(levelOrder);
	while (retval && invalid(levelOrder,tdesc))
	    retval = getNewOrder(levelOrder);
	return retval;
    }
    boolean invalid(int[] levelOrder, TermDesc tdesc) {
	for (int i = 0; i < levelOrder.length; i++) {
	    for (int j = i+1; j < levelOrder.length; j++) {
		if ((tdesc.desc(levelOrder[i]).dim() == tdesc.desc(levelOrder[j]).dim()) && 
		    (tdesc.desc(levelOrder[i]).level() < tdesc.desc(levelOrder[j]).level())) {
		    return true;
		}
	    }
	}
	return false;
    }
    boolean getNewOrder(int[] levelOrder) {
	// generate the next permutation.
	// find the rightmost value that can be incremented 
	int max = levelOrder.length-1;
	for (int i = levelOrder.length-2; i >= 0; i--) {
	    for (int val = levelOrder[i] + 1;val <= max;val++) {
		for (int j = i+1; j < levelOrder.length; j++) {
		    if (val == levelOrder[j]) {
			levelOrder[j] = levelOrder[i];
			levelOrder[i] = val;
			// reverse the values to the right of i
			int last = levelOrder.length-1;
			for (int k = i+1; k < last; k++,last--) {
			    if (levelOrder[k] > levelOrder[last]) {
				int tmp = levelOrder[k];
				levelOrder[k] = levelOrder[last];
				levelOrder[last] = tmp;
			    }
			}
			return true;
		    }
		}
	    }
	}
	return false;
    }
    public FindDiff(IterGen itergen, DiffParams p, CoverStatus status) {
	try {
	if (! p.exhaustive()) {
	    getDiff(itergen.getIter(null), p, status, null);
	    finalRes = lastAgg.finalResult();
	    allFinalRes = lastAgg.allFinalResults();
	    return;
	} 
 	int[] levelOrder = new int[itergen.termDesc().size()];
	double minCost = Double.MAX_VALUE;
	int minSz = 0;
	levelOrder[0] = -1;
	while (getNewOrder(levelOrder,itergen.termDesc())) {
	    System.out.print("Current Order: ");
	    for (int k = 0; k < levelOrder.length; k++) {
		System.out.print(levelOrder[k] + " ");
	    }
	    System.out.println("");
	    getDiff(itergen.getIter(levelOrder), p, status, levelOrder);
	    DiffResult res = lastAgg.finalResult();
	    // TODO check the result and decide on the next permutation..
	    // 
	    double cost = res.cost();
	    int sz = res.numTups();
	    if ((minCost > cost) || ((minCost == cost) && (minSz > sz))){
		minCost = cost;
		finalRes = res;
		minSz = sz;
		allFinalRes = lastAgg.allFinalResults();
	    }
	    String st = " ";
	    for (int j = 0; j < levelOrder.length; st += (levelOrder[j++] + " "));
	    System.out.println("Diff cost: " + cost + st );
	    res.output();

	    System.out.println("Min diff cost so far: " + minCost);
	    finalRes.output();
	}
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }
};


